import logging

import click

from decentnet.cli.client import client
from decentnet.cli.keys import key
from decentnet.cli.service import service
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


cli.add_command(key)
cli.add_command(service)
cli.add_command(client)

if __name__ == '__main__':
    setup_logger(True, logger)
    cli()


def main():
    setup_logger(True, logger)
    cli()



import argon2

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.modules.pow.difficulty import Difficulty


def hash_func(data: bytes, diff: Difficulty):
    return argon2.hash_password_raw(data, BlockchainParams.default_salt, diff.t_cost,
                                    diff.m_cost,
                                    diff.p_cost, diff.hash_len_chars)



import logging

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.routing_params import DEFAULT_CAPACITY
from decentnet.modules.forwarding.flow_net import FlowNetwork
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class ProcessingBlock:
    @staticmethod
    def proces_broadcast_block(network: FlowNetwork, data: dict):
        logger.debug(
            f"Adding edge from broadcast data {data['pub']} => {data['target']}")
        network.add_edge(data["pub"], data["target"], DEFAULT_CAPACITY)



"""Block Byte allocation for serialization and deserialization"""
from decentnet.__version__ import network_version

# This blocks maximum size is intentionally smaller
# to prevent DDOS attacks and ensure decentralized sending
MAXIMUM_BLOCK_SIZE = 4194304  # Bytes
BLOCK_PREFIX_LENGTH_BYTES = 3  # Bytes
INDEX_SIZE = 5  # Bytes, It would take 100 years at 10ms a block to overflow
VERSION_SIZE = len(network_version)
NONCE_SIZE = 4  # Bytes, Standard nonce size
TIMESTAMP_SIZE = 8  # Bytes, Sufficient for 2.28 million years.

MERGED_DIFFICULTY_BYTE_LEN = 13  # Bytes
HASH_LEN = 32  # Bytes
HASH_LEN_BLOCK = 48  # Bytes

RESERVED_EMPTY_BYTE_LEN = 77  # Bytes

MAXIMUM_DATA_SIZE = (MAXIMUM_BLOCK_SIZE - BLOCK_PREFIX_LENGTH_BYTES - INDEX_SIZE -
                     VERSION_SIZE - NONCE_SIZE - TIMESTAMP_SIZE -
                     MERGED_DIFFICULTY_BYTE_LEN - HASH_LEN -
                     HASH_LEN_BLOCK - RESERVED_EMPTY_BYTE_LEN)



import random
import unittest
from random import randint as ri

import rich

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.modules.blockchain.blockchain import Blockchain
from decentnet.modules.pow.difficulty import Difficulty


class BlockchainTests(unittest.TestCase):
    def test_init(self):
        blockchain: Blockchain = Blockchain("some_msg")
        seed = blockchain.chain[0]
        print(f"Succeed {seed.compute_hash().value_as_hex()}")

    def test_insertion(self):
        blockchain: Blockchain = Blockchain("some_msg")
        seed = blockchain.chain[0]
        testdata = b"testssss"

        try:
            b = blockchain.template_next_block(BlockchainParams.low_diff, b"aqweqe")
            b.mine()
            self.assertTrue(blockchain.insert(b))
        except Exception as ex:
            self.fail(f"Thrown {ex}")
        self.assertEqual(len(blockchain.chain), 2)

    def test_insertion_long(self):
        blockchain: Blockchain = Blockchain("some_msg")

        b_len = 50000

        for i in range(b_len):
            testdata = random.randbytes(8)
            try:
                b = blockchain.template_next_block(BlockchainParams.low_diff, testdata)
                b.mine()
                print(f"Mined block {b} {b.ttc} ms")
                self.assertTrue(blockchain.insert(b))
            except Exception as ex:
                self.fail(f"Thrown {ex}")
        self.assertEqual(len(blockchain.chain), b_len + 1)

    def test_insertion_very_long_intensive(self):
        rich.print("[red]This test can take hours to run[/red]")
        blockchain: Blockchain = Blockchain("veeery long test")

        b_len = 50000

        for i in range(b_len):
            testdata = random.randbytes(random.randint(1, 5000))
            diff = Difficulty(ri(1, 8), BlockchainParams.low_diff.m_cost,
                              BlockchainParams.low_diff.p_cost, ri(1, 10),
                              BlockchainParams.low_diff.hash_len_chars)
            blockchain.difficulty = diff
            b = blockchain.template_next_block(diff, testdata)
            b.mine()
            self.assertTrue(blockchain.insert(b))

        self.assertEqual(len(blockchain.chain), b_len + 1)


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
from typing import Optional, Tuple

import cbor2
from cbor2 import CBORError

from decentnet.consensus.beam2beam import HANDSHAKE_MSG
from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.interface_constants import DEFAULT_ACK_MSG_SIGN
from decentnet.consensus.net_constants import SEED_NODES_IPV6, SEED_NODES_IPV4
from decentnet.modules.comm.beacon import Beacon
from decentnet.modules.comm.beam import Beam
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.forwarding.flow_net import FlowNetwork
from decentnet.modules.key_util.key_manager import KeyManager
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.difficulty import Difficulty

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class BasicInterface:
    """
    The BasicInterface class provides methods to manage encrypted communication channels
    using beacon and beam objects. It supports initialization of sending and receiving
    channels, and handles encrypted data transmission over a decentralized network.
    """

    def __init__(self, target_key: str, target_enc_key: str):
        """
        Initialize the BasicInterface with target keys.

        Args:
            target_key (str): The public key of the target.
            target_enc_key (str): The encryption key of the target.
        """
        self.pub_key = None
        self._beam_receive = None
        self._beacon_receive = None
        self._beam_send = None
        self._beacon_send = None
        self._target_key = target_key
        self._target_enc_key = target_enc_key
        self._network = FlowNetwork()
        self.connected_relays = []

        self.encrypted_send = False
        self.encrypted_receive = False

    def init_sending(self, owner_sign_key_send_id: int, owner_enc_key_send_id: int):
        """
        Initialize the sending communication by connecting to a relay and setting up encryption.

        Args:
            owner_sign_key_send_id: The signing key ID for sending.
            owner_enc_key_send_id: The encryption key ID for sending.

        Returns:
            bool: True if encryption initialization is complete, False otherwise.
        """
        beacon_a, beam_a = self._single_relay_connect_outgoing(owner_sign_key_send_id, owner_enc_key_send_id)

        self._beacon_send = beacon_a
        self._beam_send = beam_a
        self.encrypted_send = self._beam_send.encryption_init_complete
        self.pub_key = self._beacon_send.pub_key

        return self._beam_send.encryption_init_complete

    def init_receiving(self, owner_sign_key_receive_id: int, owner_enc_key_receive_id: int):
        """
        Initialize the receiving communication by connecting to a relay and setting up encryption.

        Args:
            owner_sign_key_receive_id (int): The signing key ID for receiving.
            owner_enc_key_receive_id (int): The encryption key ID for receiving.

        Returns:
            bool: True if encryption initialization is complete, False otherwise.
        """
        beacon_b, beam_b = self._single_relay_connect_incoming(owner_sign_key_receive_id,
                                                               owner_enc_key_receive_id)

        self._beacon_receive = beacon_b
        self._beam_receive = beam_b
        self.encrypted_receive = self._beam_receive.encryption_init_complete

        return self._beam_receive.encryption_init_complete

    def _single_relay_connect_outgoing(self, owner_key_id: int, owner_enc_key_id: int) -> Tuple[
        Beacon, Beam]:
        """
        Connect to a single relay and initialize encryption for outgoing encryption.

        Args:
            owner_key_id (int): The owner's signing key ID.
            owner_enc_key_id (int): The owner's encryption key ID.

        Returns:
            Tuple[Beacon, Beam]: Connected Beacon and Beam objects.
        """
        beacon, beam = self.__auto_select_single_seed_node(owner_key_id, owner_enc_key_id)
        if not beacon:
            logger.critical("No connectable relay found.")
            exit(1)
        self._init_out_encryption(beam)
        return beacon, beam

    def _single_relay_connect_incoming(self, owner_key_id: int, owner_enc_key_id: int) -> Tuple[
        Beacon, Beam]:
        """
        Connect to a single relay and initialize encryption for incoming connections.

        Args:
            owner_key_id (int): The owner's signing key ID.
            owner_enc_key_id (int): The owner's encryption key ID.

        Returns:
            Tuple[Beacon, Beam]: Connected Beacon and Beam objects.
        """
        beacon, beam = self.__auto_select_single_seed_node(owner_key_id, owner_enc_key_id)
        if not beacon:
            logger.critical("No connectable relay found.")
            exit(1)

        return beacon, beam

    def __auto_select_single_seed_node(self, owner_key_id: int, owner_enc_key_id: int):
        """
        Automatically select and connect to a seed node.

        Args:
            owner_key_id (int): The owner's signing key ID.
            owner_enc_key_id (int): The owner's encryption key ID.

        Returns:
            Tuple[Optional[Beacon], Optional[Beam]]: Connected Beacon and Beam objects if successful,
            otherwise (False, None).
        """
        seed_nodes = SEED_NODES_IPV4 + SEED_NODES_IPV6

        for node_v4 in seed_nodes:
            logger.debug(f"Trying to connect to {node_v4[0]}:{node_v4[1]}")
            node_v4_concat = "".join([str(n) for n in node_v4])
            if node_v4_concat + str(owner_key_id) not in self.connected_relays:
                try:
                    beacon = Beacon(node_v4[0], node_v4[1], owner_key_id)
                    beam = beacon.create_beam(self._target_key, owner_enc_key_id, loopback=False)

                    if beam is None:
                        continue

                    _, o_pub_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(beam.pub_key_id))
                    pub_key = AsymCrypt.verifying_key_to_string(o_pub_key)

                    self._network.add_edge(pub_key,
                                           beam.target_key, 10)

                    logger.info(f"[OK] Seed {node_v4[0]}:{node_v4[1]} connected!")
                    self.connected_relays.append(node_v4_concat + str(owner_key_id))
                    return beacon, beam
                except (ConnectionError, ConnectionResetError, ConnectionRefusedError,
                        ConnectionAbortedError, TimeoutError):
                    logger.debug(f"{node_v4[0]}:{node_v4[1]} failed to connect")
            else:
                logger.debug(
                    "This relay is not connectable with this key pair, as different beam is already connected.")
        return False, None

    def _init_out_encryption(self, beam: Beam, difficulty: Difficulty = BlockchainParams.low_diff):
        """
        Initialize encryption for a Beam.

        Args:
            beam (Beam): The Beam object to initialize encryption on.
            difficulty (Difficulty, optional): The difficulty level for the secure communication.
            Defaults to BlockchainParams.low_diff.

        Returns:
            bool: True if encryption is successfully initialized, False otherwise.
        """
        if beam.encryption_init_complete:
            logger.debug(f"Beam {beam.hash} already initialized encryption")
            return True

        beam.set_communication_difficulty(difficulty)
        logger.debug(f"Init b2b connection with {self._target_enc_key}")
        beam.init_secure_b2b_connection(self._target_enc_key)

        logger.debug("Waiting for ack handshake")
        dec_data = self._fetch_ack(beam)
        beam.encryption_init_complete = (dec_data["status"] == "OK")
        return beam.encryption_init_complete

    @classmethod
    def _fetch_ack(cls, beam: Beam):
        """
        Fetch and decrypt the acknowledgement message from a Beam.

        Args:
            beam (Beam): The Beam object to fetch the acknowledgement from.

        Returns:
            dict: The decrypted acknowledgement data.
        """
        data = beam.fetch_message()[0].data
        dec_data = cbor2.loads(beam.encryptor_beacon.decrypt(data))
        return dec_data

    def send(self, data: bytes, difficulty: Difficulty = BlockchainParams.low_diff,
             require_ack: bool = False) -> Optional[bool]:
        """
        Send data over an initialized Beam with optional acknowledgement.

        Args:
            data (bytes): The data to be sent.
            difficulty (Difficulty, optional): The difficulty level for the secure communication.
            Defaults to BlockchainParams.low_diff.
            require_ack (bool, optional): Whether to wait for an acknowledgement.
            Defaults to False.

        Returns:
            Optional[bool]: Bool if acknowledgement is required and status is OK, otherwise None.
        """
        if self._beam_send is None or self._beacon_send is None:
            raise Exception("You need to initialize sending first!")

        self._beam_send.comm_bc.difficulty = difficulty

        e_data = data + DEFAULT_ACK_MSG_SIGN if require_ack else data

        self._beam_send.send_communication_data(e_data)

        if require_ack:
            dec_data = self._fetch_ack(self._beam_send)
            return dec_data["status"] == "OK"
        return None

    def fetch(self) -> dict | None:
        """
        Fetch and decrypt data from the initialized receiving Beam.

        Returns:
            dict: The decrypted data as a dictionary.

        Raises:
            Exception: If receiving has not been initialized.
        """
        if self._beam_receive is None or self._beacon_receive is None:
            raise Exception("You need to initialize receiving first!")

        logger.info("Waiting for incoming beams")
        enc_data = self._beam_receive.fetch_message()[0].data
        data = self._beam_receive.encryptor_beacon.decrypt(enc_data)
        if data[-4:] == b'ACK!':
            self._beam_receive.send_communication_data(HANDSHAKE_MSG)
        try:
            return cbor2.loads(data)
        except CBORError:
            return None



import unittest
from random import randint

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.modules.blockchain.block import Block


class TestBlockSerialization(unittest.TestCase):
    def setUp(self):
        self.index = 1
        self.prev_hash = bytes.fromhex('abcd' * 8)  # 32 bytes of data
        self.difficulty = BlockchainParams.low_diff  # Assume a valid Difficulty object can be instantiated like this
        self.data = bytearray(b'some data for testing')

    def test_block_to_from_bytes(self):
        # Create a block
        block = Block(self.index, self.prev_hash, self.difficulty, self.data)

        # Serialize to bytes
        block_bytes = block.to_bytes()

        # Deserialize from bytes
        deserialized_block = Block.from_bytes(block_bytes)

        # Assert that the deserialized block matches the original block
        self.assertEqual(block.index, deserialized_block.index)
        self.assertEqual(block.previous_hash, deserialized_block.previous_hash)
        self.assertEqual(block.diff, deserialized_block.diff)
        self.assertEqual(block.data, deserialized_block.data)
        self.assertEqual(block.version, deserialized_block.version)
        self.assertEqual(block.nonce, deserialized_block.nonce)
        self.assertAlmostEqual(block.timestamp, deserialized_block.timestamp, places=5)

    def test_block_to_from_bytes_prevhash(self):
        # Create a block

        for _ in range(1000):
            non_empty_bytes_by4 = randint(1, 8)
            prev_hash = bytes.fromhex("00" * (16 - non_empty_bytes_by4)) + bytes.fromhex(
                'abcd' * non_empty_bytes_by4)
            block = Block(self.index, prev_hash, self.difficulty, self.data)

            # Serialize to bytes
            block_bytes = block.to_bytes()

            # Deserialize from bytes
            deserialized_block = Block.from_bytes(block_bytes)

            # Assert that the deserialized block matches the original block
            self.assertEqual(block.index, deserialized_block.index)
            self.assertEqual(block.previous_hash, deserialized_block.previous_hash)
            self.assertEqual(block.diff, deserialized_block.diff)
            self.assertEqual(block.data, deserialized_block.data)
            self.assertEqual(block.version, deserialized_block.version)
            self.assertEqual(block.nonce, deserialized_block.nonce)
            self.assertAlmostEqual(block.timestamp, deserialized_block.timestamp, places=5)


if __name__ == '__main__':
    unittest.main()



import logging

import pyximport
import rich

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.pow.pow_utils import PowUtils

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

pyximport.install()
try:
    from decentnet.modules.pow.computation_fast import compute_pow  # noqa
except ImportError as ex:
    rich.print(f"[red]POW Operating in low hashrate mode, more instructions: {ex}[/red]")
    from decentnet.modules.pow.computation_slow import compute_pow


class PoW:
    @staticmethod
    def compute(input_hash, n_bits: int, start_nonce=0):
        nonce = compute_pow(n_bits, input_hash, start_nonce)
        check_b = PowUtils.get_bit_length(input_hash, nonce, input_hash.diff)
        return input_hash, nonce



import unittest

from decentnet.modules.key_util.key_manager import KeyManager


class TestSSHKeyManager(unittest.TestCase):
    def setUp(self):
        self.key_manager = KeyManager()

    def test_generate_ssh_key_pair(self):
        private_key, public_key = self.key_manager.generate_singing_key_pair()
        self.assertIsNotNone(private_key)
        self.assertIsNotNone(public_key)

    def test_save_and_retrieve_ssh_key_pair(self):
        private_key, public_key = self.key_manager.generate_singing_key_pair()
        _id = self.key_manager.save_to_db(private_key, public_key, "test",
                                          can_encrypt=False)

        retrieved_private_key, retrieved_public_key = self.key_manager.retrieve_ssh_key_pair_from_db(
            _id)
        self.assertEqual(private_key, retrieved_private_key)
        self.assertEqual(public_key, retrieved_public_key)



#cython: language_level=3
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class AESCipher:
    def __init__(self, password: bytes, key_size: int = 256, salt: bytes = None):
        if key_size not in (128, 192, 256):
            raise ValueError("Key size must be 128, 192, or 256 bits.")
        self.password = password
        self.key_size = key_size
        self.salt = salt if salt is not None else os.urandom(16)
        self.backend = default_backend()

    def derive_key(self, password: bytes, salt: bytes, key_size: int):
        """Derive a cryptographic key from the given password and salt."""
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=key_size // 8,
            salt=salt,
            iterations=100000,
            backend=self.backend
        )
        return kdf.derive(password)

    def encrypt(self, plaintext: bytes) -> bytes:
        key = self.derive_key(self.password, self.salt, self.key_size)
        iv = os.urandom(16)
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv), self.backend)
        encryptor = cipher.encryptor()
        padder = padding.PKCS7(128).padder()
        padded_data = padder.update(plaintext) + padder.finalize()
        cipher_text = encryptor.update(padded_data) + encryptor.finalize()
        return self.salt + iv + cipher_text  # Prepend the salt and IV to the ciphertext

    def decrypt(self, cipher_text: bytes) -> bytes:
        salt = cipher_text[:16]  # Extract the salt
        iv = cipher_text[16:32]  # Extract the IV
        key = self.derive_key(self.password, salt,
                              self.key_size)  # Re-derive the key using the extracted salt
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv), self.backend)
        decryptor = cipher.decryptor()
        padded_plaintext = decryptor.update(cipher_text[32:]) + decryptor.finalize()
        unpadder = padding.PKCS7(128).unpadder()
        plaintext = unpadder.update(padded_plaintext) + unpadder.finalize()
        return plaintext
